"""View module."""
from .export import Export
from .wait import Wait
from .widget import Labeled, VariableContainer, AdjustableInput
from .view import (
    StateInteraction,
    FileInteraction,
    ColorInteraction,
    IterationInteraction,
    PositioningInteraction,
    View
)

__all__ = [
    "Export", "StateInteraction", "FileInteraction", "ColorInteraction",
    "IterationInteraction", "PositioningInteraction", "View", "Wait",
    "Labeled", "VariableContainer", "AdjustableInput"
]



"""Model module."""
from .fractale import ModuloColoration, Fractale, Julia, Mandelbrot
from .manager import FractaleManager, DataExport


__all__ = [
    "ModuloColoration", "Fractale", "Julia",
    "Mandelbrot", "FractaleManager", "DataExport"
]



"""Unit tests for mandelia.model."""
from unittest import TestCase
from mandelia.model import Mandelbrot, Julia, ModuloColoration


class TestMandelbrot(TestCase):

    def setUp(self) -> None:
        self.color = ModuloColoration(9, 2, 3)
        self.mandelbrot = Mandelbrot(self.color, width=256, height=128)

    def test_image(self) -> None:
        img = self.mandelbrot.image()
        w, h = img.size
        self.assertEqual(w, 256)
        self.assertEqual(h, 128)


class TestJulia(TestCase):

    def setUp(self) -> None:
        self.color = ModuloColoration(9, 2, 3)
        self.julia = Julia(self.color, width=256, height=128)

    def test_image(self) -> None:
        img = self.julia.image()
        w, h = img.size
        self.assertEqual(w, 256)
        self.assertEqual(h, 128)



"""Contains reuseable widgets."""
import tkinter as tk
from abc import ABC
from tkinter.constants import LEFT, NW, HORIZONTAL, FALSE, X, SUNKEN
from typing import Generic, Optional, TypeVar

T = TypeVar("T", tk.StringVar, tk.IntVar, tk.DoubleVar, tk.BooleanVar)


class Labeled(tk.Frame, ABC):
    """Label for another widget."""
    def __init__(self, master: Optional[tk.Misc], text: str) -> None:
        """Instantiate Labeled."""
        super().__init__(master)
        self.label = tk.Label(self, text=text, width=10, anchor=NW)
        self.label.pack(side=LEFT)


class VariableContainer(Generic[T]):
    """Contain a variable."""
    def __init__(self, var: T) -> None:
        """Instantiate Labeled."""
        self.__var: T = var

    @property
    def var(self) -> T:
        """Get the variable."""
        return self.__var


class AdjustableInput(Labeled, VariableContainer[tk.DoubleVar]):
    """Widget to have a variable with a scale and an entry."""
    def __init__(self, master: Optional[tk.Misc], text: str, min_: int,
                 default: int, max_: int, entry_width: int = 6) -> None:
        """Instantiate AdjustableInput."""
        Labeled.__init__(self, master, text)
        VariableContainer.__init__(self, tk.DoubleVar(self, float(default)))
        self.min = min_
        self.max = max_
        self.default = default
        self.__error = False
        self.entry_var = tk.StringVar(self, str(self.default))
        self.scale = tk.Scale(self, from_=self.min, to=self.max,
                              sliderlength=30, orient=HORIZONTAL,
                              variable=self.var, showvalue=FALSE)
        self.entry = tk.Entry(self, width=entry_width,
                              textvariable=self.entry_var)
        self.scale.pack(side=LEFT, fill=X, expand=True)
        self.entry.pack(side=LEFT, padx=3)
        self.entry_var.trace_add("write", self.on_update_scale)
        self.var.trace_add("write", self.on_update_entry)
        self.entry.bind("<FocusOut>", self._event_focus_out)
        self.entry.bind("<KeyPress>", self.on_keypress)

    def on_keypress(self, event):
        # type: (tk.Event[tk.Entry]) -> None
        """Handle keypress"""
        delta = {"Down": -1, "Up": 1}.get(event.keysym, 0)
        if delta:
            self.var.set(max(self.var.get() + delta, self.min))

    def disable(self) -> None:
        """Disable the entry."""
        self.entry["state"] = "disable"
        self.scale["sliderrelief"] = "sunken"
        self.scale["state"] = "disable"
        self.label["state"] = "disable"

    def enable(self) -> None:
        """Enable the entry."""
        self.entry["state"] = "normal"
        self.scale["sliderrelief"] = "raised"
        self.scale["state"] = "normal"
        self.label["state"] = "normal"

    @property
    def error(self) -> bool:
        """Get the error state."""
        return self.__error

    @error.setter
    def error(self, state: bool) -> None:
        """Set error on true or false."""
        self.__error = bool(state)
        self.entry.configure(bg="red" if self.__error else "white")

    def on_update_scale(self, name: str, index: str, mode: str) -> None:
        """Called for update the scale."""
        text = self.entry_var.get()
        self.error = False
        max_chars = max(len(str(self.max)), 1)
        nb_chars = len(self.entry_var.get())
        if nb_chars > max_chars:
            text = text[:max_chars]
            self.entry_var.set(text)

        try:
            value = int(text)
        except ValueError:
            self.error = True
        else:
            if self.min > value:
                self.error = True
            elif self.max < value:
                self.entry_var.set(str(self.max))
                self.scale.set(self.max)  # type: ignore
            else:
                self.scale.set(value)  # type: ignore

    def on_update_entry(self, name: str, index: str, mode: str) -> None:
        """Called for update the entry."""
        self.entry_var.set(str(int(self.var.get())))

    def _event_focus_out(self, event):
        # type: (tk.Event[tk.Entry]) -> None
        """Return to the previous valid state."""
        if self.error:
            try:
                int(self.entry_var.get())
            except ValueError:
                min_chars = max(len(str(self.min)), 1)
                nb_chars = len(self.entry_var.get())
                if nb_chars < min_chars:
                    self.scale.set(str(self.min))  # type: ignore
            self.on_update_entry(name="", index="", mode="")


class Output(Labeled, VariableContainer[tk.StringVar]):
    """Read-Only Entry."""
    def __init__(self, master: Optional[tk.Misc],
                 text: str, default: str) -> None:
        """Instantiate Output."""
        Labeled.__init__(self, master, text)
        VariableContainer.__init__(self, tk.StringVar(self, default))
        self.sep = tk.Label(self, text=":")
        self.output_label = tk.Label(self, relief=SUNKEN, anchor=NW,
                                     textvariable=self.var, width=25)
        self.sep.pack(side=LEFT)
        self.output_label.pack(side=LEFT, fill=X, expand=True, padx=2, pady=2)


class InputOutput(Labeled, VariableContainer[tk.StringVar]):
    """Writable Entry."""
    def __init__(
        self, master: Optional[tk.Misc], text: str, default: str
    ) -> None:
        """Instantiate InputOutput."""
        Labeled.__init__(self, master, text)
        VariableContainer.__init__(self, tk.StringVar(self, default))
        self.sep = tk.Label(self, text=":")
        self.entry = tk.Entry(self, textvariable=self.var, width=32)
        self.sep.pack(side=LEFT)
        self.entry.pack(side=LEFT, fill=X, padx=3)



# distutils: language=c++

import numpy as np
import struct as s
import cv2
from math import log
from PIL import Image

cimport numpy as np
cimport cython

from cython.parallel import prange

DTYPE = np.uint32
ctypedef np.uint32_t DTYPE_t

COLORTYPE = np.uint8
ctypedef np.uint8_t COLORTYPE_t

DEF PIXEL_DEFAULT = 0.02
DEF MIN_PIXEL_SIZE = PIXEL_DEFAULT * 16


cdef unsigned int iterate(double z_r, double z_i, double c_r, double c_i,
                          unsigned int iterations) nogil except +:
    """Iterate on a complex numbers. formula: Zn+1 = Zn ** 2 + C"""
    cdef:
        double tmp
        unsigned int i

    for i in range(iterations):
        tmp = z_r
        z_r = z_r * z_r - z_i * z_i + c_r
        z_i = 2 * z_i * tmp + c_i
        if z_r * z_r + z_i * z_i > 4:
            break
    return 0 if i == iterations - 1 else i + 1


modulo_coloration_saver = s.Struct("BBB")
cdef class ModuloColoration:
    cdef:
        public unsigned char r, g, b

    def __init__(self, r=3, g=1, b=10):
        self.r, self.g, self.b = r, g, b

    cpdef to_bytes(self):
        """Convert ModuloColor into bytes."""
        return modulo_coloration_saver.pack(self.r, self.g, self.b)

    cpdef from_bytes(self, bytes bytes_):
        """Convert bytes into ModuloColor."""
        self.r, self.g, self.b = modulo_coloration_saver.unpack(bytes_)

    cpdef bytes_size(self):
        return modulo_coloration_saver.size

    @cython.boundscheck(False)  # turn off bounds-checking
    @cython.wraparound(False)  # turn off negative index wrapping
    cpdef np.ndarray[COLORTYPE_t, ndim=3] colorize(
            self, np.ndarray[DTYPE_t, ndim=2] np_fractale):
        """ColorInteraction a two-dimensional array."""
        cdef:
            short width, height, x, y
            unsigned char r = self.r, g = self.g, b = self.b
            unsigned int i
            np.ndarray[COLORTYPE_t, ndim=3] image

        width = np_fractale.shape[0]
        height = np_fractale.shape[1]
        image = np.zeros((height, width, 3), dtype=COLORTYPE)
        for x in prange(width, schedule='guided', nogil=True):
            for y in range(height):
                i = np_fractale[x, y]
                if i != 0:
                    image[y, x, 0] = r * i
                    image[y, x, 1] = g * i
                    image[y, x, 2] = b * i
        return image

fractale_saver = s.Struct("dddhhI")
cdef class Fractale:
    cdef:
        readonly double real, imaginary, pixel_size
        readonly short width, height
        readonly unsigned int iterations
        readonly need_update
        content
        ModuloColoration color

    def __init__(self, ModuloColoration color, real=0.0, imaginary=0.0,
                 iterations=1_000, width=256, height=256,
                 pixel_size=PIXEL_DEFAULT):
        self.content = np.zeros((width, height), dtype=DTYPE)
        self.color = color
        self.real = real
        self.imaginary = imaginary
        self.iterations = iterations
        self.width = width
        self.height = height
        self.pixel_size = pixel_size
        self.need_update = True

    def __copy__(self):
        data = self.to_bytes()
        color = type(self.color)()
        frac = type(self)(color)
        frac.from_bytes(data)
        return frac

    def drop(self, metadata, handler_progress = None):
        # TODO safe check
        if handler_progress is None:
            handler_progress = lambda *args, **kwargs: None
        path: str = metadata["path"]
        ext: str = metadata.get("ext", path.lower().rsplit(".", 1)[-1])
        width, height = metadata["width"], metadata["height"]
        compression = metadata["compression"]
        fractale = self.__copy__()
        open(metadata["path"], "a").close()  # test writable
        if ext in ("png", "pns", "jpeg", "jpe", "jpeg"):
            img: Image.Image = fractale.image_at_size(width, height)
            img.save(path, quality=compression)
            handler_progress(1, img)
        elif ext in ("gif", "mp4"):
            pixel_size = fractale.pixel_size
            speed = metadata["speed"] / 10
            fractale.top()
            fractale.resize(width, height)
            img = fractale.image()
            handler_progress(0, img)
            fps = metadata["fps"]
            multi = 1 + speed / fps
            start_pixel_size = fractale.pixel_size
            s = log(start_pixel_size, multi)
            e = log(pixel_size, multi)

            def iterate_images():
                progression = 0
                yield fractale.image()
                while fractale.pixel_size > pixel_size:
                    fractale.middle_zoom(multi)
                    p = log(fractale.pixel_size, multi)
                    progression = max(0, min(((p - s) / (e - s), 1)))
                    print(f"Progress : {progression * 100:.2f}%")
                    img = fractale.image()
                    handler_progress(progression, img)
                    yield img
                if progression != 1:
                    handler_progress(1, fractale.image())

            if ext == "gif":
                img.save(metadata["path"], fromat='GIF', save_all=True,
                         append_images=list(iterate_images()),
                         optimize=True, duration=int(1000 / fps),
                         loop=0, palette=Image.ADAPTIVE, disposal=1)
            else:
                codec = cv2.VideoWriter_fourcc(*'mp4v')
                video = cv2.VideoWriter(path, codec, fps, (width, height))
                video.write(cv2.cvtColor(np.array(img),  # type: ignore
                                         cv2.COLOR_RGB2BGR))
                for frame in iterate_images():
                    video.write(cv2.cvtColor(np.array(frame),  # type: ignore
                                             cv2.COLOR_RGB2BGR))
                video.release()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return (f"<{self.__class__.__name__} "
                f"pixel_size={self.pixel_size} "
                f"{self.real}{self.imaginary:+}i>")

    cpdef set_real(self, double real):
        """Set real part of Z."""
        self.real = real
        self.need_update = True

    cpdef set_imaginary(self, double imaginary):
        """Set imaginary part of Z."""
        self.imaginary = imaginary
        self.need_update = True

    cpdef set_iterations(self, unsigned int iterations):
        """Set max iterations."""
        self.iterations = iterations

    cpdef resize(self, short width, short height):
        """Resize width and height, and adjust the zoom if necessary."""
        cdef:
            short tmp_width
            double multiplier
        tmp_width = self.width
        self.width = width
        self.height = height
        try:
            self.middle_zoom(width / tmp_width)
        except ZeroDivisionError:
            pass
        self.need_update = True

    cpdef iterations_sum(self):
        """Total number of iterations."""
        cdef unsigned long long sum_
        sum_ = np.sum(self.content)
        sum_ += np.count_nonzero(self.content == 0) * self.iterations
        return sum_

    cpdef iterations_per_pixel(self):
        """Average iteration per pixel."""
        cdef unsigned int pixels
        pixels = self.width * self.height
        return self.iterations_sum() / pixels if pixels != 0 else 0

    cdef _check_min_size(self):
        """Increases pixel size if it is smaller than MIN_PIXEL_SIZE."""
        cdef double multiplier = self.pixel_size / MIN_PIXEL_SIZE
        if self.pixel_size > MIN_PIXEL_SIZE:
            self.pixel_size = MIN_PIXEL_SIZE
            self.real = self.real / multiplier
            self.imaginary = self.imaginary / multiplier
            if -0.001 < self.real < 0.001:
                self.real = 0
            if -0.001 < self.imaginary < 0.001:
                self.imaginary = 0

    cpdef image(self):
        """
        Compute image if there is update otherwise just do the coloring.
        """
        cdef np.ndarray[COLORTYPE_t, ndim=3] colored
        if self.need_update:
            self._compute()
            self.need_update = False
        colored = self.color.colorize(self.content)
        return Image.fromarray(colored, 'RGB')

    cpdef image_at_size(self, short width, short height):
        """
        Get image with specific size.
        """
        cdef:
            np.ndarray[DTYPE_t, ndim=2] content_copy
            short x
            short w_copy = self.width
            short h_copy = self.height
            double real_copy = self.real
            double imaginary_copy = self.imaginary
            double pixel_copy = self.pixel_size

        if w_copy != width or h_copy != height:
            content_copy = self.content.copy()
            self.resize(width, height)
            img = self.image()
            self.real = real_copy
            self.imaginary = imaginary_copy
            self.pixel_size = pixel_copy
            self.content = content_copy
            self.width = w_copy
            self.height = h_copy
        else:
            img = self.image()
        return img

    cpdef top(self):
        """Set pixel at default size."""
        w, h = self.width, self.height
        self.resize(256, 256)
        self.pixel_size = PIXEL_DEFAULT
        self.resize(w, h)

    cpdef middle_zoom(self, double multiplier):
        """Zoom at the middle."""
        self.zoom(self.width // 2, self.height // 2, multiplier)

    cpdef zoom(self, short x, short y, double multiplier):
        """Zoom at a position in the image."""
        cdef double pixel = self.pixel_size, real = self.real, imaginary = self.imaginary
        if -1e-09 < multiplier < 1e-09:
            raise ZeroDivisionError(
                f"the multiplier is too close to zero ({multiplier})")
        self.real = (real * 1 / multiplier
                     + (real + x * pixel - self.width / 2 * pixel)
                     * (1 - 1 / multiplier))
        self.imaginary = (imaginary * 1 / multiplier
                          + (imaginary + y * pixel - self.height / 2 * pixel)
                          * (1 - 1 / multiplier))
        self.pixel_size /= multiplier
        self._check_min_size()
        self.need_update = True

    cpdef reset(self):
        """Reset position and pixel size."""
        self.real = 0
        self.imaginary = 0
        self.pixel_size = PIXEL_DEFAULT
        self.need_update = True

    cdef np.ndarray[DTYPE_t, ndim=2] _compute(self):
        """Compute fractale."""
        raise NotImplementedError()

    cpdef real_at_x(self, short x):
        """Return real part of Z at x in image."""
        cdef double start_r = self.real - (self.width * self.pixel_size) / 2
        return start_r + x * self.pixel_size

    cpdef imaginary_at_y(self, short y):
        """Return imaginary part of Z at y in image."""
        cdef double start_i = self.imaginary - (
                self.height * self.pixel_size) / 2
        return start_i + y * self.pixel_size

    cpdef to_bytes(self):
        """Return bytes representative of the fractal."""
        data = fractale_saver.pack(
            self.real,
            self.imaginary,
            self.pixel_size,
            self.width,
            self.height,
            self.iterations
        )
        data += self.color.to_bytes()
        return data

    cpdef from_bytes(self, bytes bytes_):
        """Load data on the fractal."""
        cdef short w = self.width, h = self.height
        record = fractale_saver.unpack(bytes_[:32])
        (self.real, self.imaginary, self.pixel_size,
            self.width, self.height, self.iterations) = record
        self.resize(w, h)
        self.color.from_bytes(bytes_[32:])
        self.need_update = True

    cpdef bytes_size(self):
        return fractale_saver.size + self.color.bytes_size()


julia_saver = s.Struct("dd")
cdef class Julia(Fractale):
    cdef:
        readonly double c_r, c_i
    def __init__(self, color: ModuloColoration, c_r=0, c_i=0, real=0,
                 imaginary=0, iterations=1_000, width=48, height=48,
                 pixel_size=PIXEL_DEFAULT):
        self.c_r = c_r
        self.c_i = c_i
        super().__init__(color, real, imaginary, iterations, width,
                         height, pixel_size)

    cpdef set_c_r(self, double c_r):
        """Set real part of C."""
        self.c_r = c_r
        self.need_update = True

    cpdef set_c_i(self, double c_i):
        """Set imaginary part of C."""
        self.c_i = c_i
        self.need_update = True

    cpdef to_bytes(self):
        """Return bytes representative of the fractal."""
        return (super(Julia, self).to_bytes()
                + julia_saver.pack(self.c_r, self.c_i))

    cpdef from_bytes(self, bytes bytes_):
        """Load data on the fractal."""
        super(Julia, self).from_bytes(bytes_[:-16])
        self.c_r, self.c_i = julia_saver.unpack(bytes_[-16:])

    cpdef bytes_size(self):
        return super(Julia, self).bytes_size() + julia_saver.size

    @cython.boundscheck(False)  # turn off bounds-checking
    @cython.wraparound(False)  # turn off negative index wrapping
    cdef np.ndarray[DTYPE_t, ndim=2] _compute(self):
        """Compute fractal."""
        cdef:
            double z_r, z_i, tmp, x_start, y_start
            double real = self.real
            double imaginary = self.imaginary
            double pixel_size = self.pixel_size
            double c_r = self.c_r
            double c_i = self.c_i
            short width = self.width, height = self.height
            short y, x
            unsigned int i, iterations = self.iterations
            np.ndarray[DTYPE_t, ndim=2] content = np.zeros((width, height),
                                                           dtype=DTYPE)
        x_start = real - (width >> 1) * pixel_size
        y_start = imaginary - (height >> 1) * pixel_size
        for x in prange(width, schedule='guided', nogil=True):
            z_r = x_start + x * pixel_size
            for y in range(height):
                z_i = y_start + y * pixel_size
                content[x, y] = iterate(z_r, z_i, c_r, c_i, iterations)
        self.content = content


cdef class Mandelbrot(Fractale):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @cython.boundscheck(False)  # turn off bounds-checking
    @cython.wraparound(False)  # turn off negative index wrapping
    cdef np.ndarray[DTYPE_t, ndim=2] _compute(self):
        """Compute mandelbrot fractale."""
        cdef:
            double c_r, c_i, z_r, z_i, tmp, x_start, y_start
            double real = self.real
            double imaginary = self.imaginary
            double pixel_size = self.pixel_size
            short y, x
            unsigned int i, iterations = self.iterations
            short width = self.width, height = self.height
            np.ndarray[DTYPE_t, ndim=2] content = np.zeros((width, height),
                                                           dtype=DTYPE)
        x_start = real - (width >> 1) * pixel_size
        y_start = imaginary - (height >> 1) * pixel_size
        for x in prange(width, schedule='guided', nogil=True):
            c_r = x_start + x * pixel_size
            for y in range(height):
                c_i = y_start + y * pixel_size
                content[x, y] = iterate(0, 0, c_r, c_i, iterations)
        self.content = content



"""Wait Toplevel widget."""
from math import ceil
import tkinter as tk
from tkinter import ttk
from tkinter.constants import NW, X, HORIZONTAL, BOTTOM
from typing import Optional

from PIL import Image, ImageTk

from ..util import set_icon


class Wait(tk.Toplevel):
    """Window for preview."""
    def __init__(self, view: tk.Misc) -> None:
        """Instantiate preview window."""
        super().__init__(view)
        self.resizable(width=False, height=False)
        self.title("Opération lente")
        self.root = view
        self.configure()
        self.geometry("300x300")
        set_icon(self)
        self.__image: Optional[Image.Image] = None
        self.__image_tk: Optional[ImageTk.PhotoImage] = None
        self.__index: Optional[int] = None
        self.var = tk.IntVar(self, 0)

        self.label = tk.Label(self, text="Une opération lente est en cours")
        self.canvas = tk.Canvas(self, width=200, height=200, bd=0,
                                highlightthickness=0, bg="#c8c8c8")
        self.progressbar = ttk.Progressbar(
            self, variable=self.var, orient=HORIZONTAL, value=0, maximum=100
        )
        self.cancel = tk.Button(self, text="Annuler",
                                command=self.destroy, width=20)

        self.label.pack(fill=X, padx=20)
        self.canvas.pack()
        self.progressbar.pack(fill=X, side=BOTTOM)
        self.cancel.pack(pady=10)

        # self.take_control()

    def done(self) -> None:
        """Alias for destroy."""
        self.destroy()
        self.update()

    def set_preview(self, image: Image.Image) -> None:
        """Set image to preview."""
        width, height = image.size
        ratio_w = width / 200
        ratio_h = height / 200
        ratio = max(ratio_h, ratio_w)
        img_w = ceil(width / ratio)
        img_h = ceil(height / ratio)
        self.__image = image.resize((img_w, img_h))
        self.__image_tk = ImageTk.PhotoImage(self.__image)
        if self.__index is not None:
            self.canvas.delete(self.__index)
        self.__index = self.canvas.create_image(
            100 - img_w / 2, 100 - img_h / 2, image=self.__image_tk, anchor=NW)
        self.canvas.tag_lower(self.__index)
        self.update()

    def progress(self, percent: float) -> None:
        """Update progress bar."""
        if ceil(percent * 100) != self.var.get():
            self.var.set(ceil(percent * 100))
            self.root.update()



"""All utilities functions."""
import os
import sys

from datetime import datetime
from functools import wraps
from typing import Callable, TypeVar, Union, Optional
import tkinter as tk

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


T = TypeVar('T')
P = ParamSpec('P')


def rel_path(relative_path: str) -> str:
    """Get path as relative path, pyinstaller compatible."""
    meipass: Optional[str] = getattr(sys, "_MEIPASS", None)
    frozen: bool = getattr(sys, 'frozen', False)
    if meipass is not None:
        dir_path = os.path.join(meipass, "mandelia")
    elif frozen:
        dir_path = os.path.dirname(sys.executable)
    else:
        dir_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(dir_path, relative_path)


def logger(function: Callable[P, T]) -> Callable[P, T]:
    """Log function by displaying arguments."""
    @wraps(function)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        name = function.__qualname__
        str_args = ", ".join(repr(arg) for arg in args)  # type: ignore
        str_kwargs = ",".join("{}={!r}".format(k, arg)  # type: ignore
                              for k, arg in kwargs.items())  # type: ignore
        str_params = ",".join(part for part in (str_args, str_kwargs) if part)
        print("[{}] {}({})".format(datetime.now(), name, str_params))
        return function(*args, **kwargs)
    return wrapper


def sizeof_fmt(path: str) -> str:
    """Show information on a path."""
    size = float(os.path.getsize(path))
    for unit in ["", "k", "M", "G", "T", "P", "E", "Z"]:
        if abs(size) < 1000.0:
            return "{:3.1f}{}o".format(size, unit)
        size /= 1000.0
    return "{:.1f}Yo".format(size)


def stat_file(path: str) -> str:
    """Show information on a path."""
    return "Chemin : {}\nTaille : {}".format(repr(path), sizeof_fmt(path))


def set_icon(root: Union[tk.Toplevel, tk.Misc]) -> bool:
    """Set icon on tk.Misc or tk.TopLevel?"""
    try:
        icon = tk.PhotoImage(file=rel_path("view/images/logo.png"))
        root.tk.call("wm", "iconphoto", root._w, icon)  # type: ignore
        return True
    except OSError as err:
        print(err, file=sys.stderr)
        return False



"""Manage fractale mandelbrot and julia."""
import sys
from typing import Optional, Callable, Tuple

from PIL import Image

from .fractale import (
    Mandelbrot,
    Julia,
    Fractale,
    ModuloColoration
)


if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class DataExport(TypedDict):
    """Represent a metadata of media export."""
    path: str
    ext: str
    width: int
    height: int
    compression: int
    fps: int
    speed: int


RATIO = 3
ProgressHandler = Callable[[float, Image.Image], None]


class FractaleManager:
    """Manage fractale mandelbrot and julia."""
    def __init__(self, width: int, height: int) -> None:
        """Instantiate FractaleManager."""
        self.__coloration = ModuloColoration(3, 1, 10)
        self.__mandelbrot = Mandelbrot(self.__coloration,
                                       width=width,
                                       height=height)
        self.__julia = Julia(self.__coloration,
                             width=int(width / RATIO),
                             height=int(height / RATIO))
        self.first: Fractale = self.__mandelbrot
        self.second: Fractale = self.__julia

    def resize(self, width: int, height: int) -> None:
        """Resize 2 fractals."""
        self.first.resize(width, height)
        self.second.resize(int(width / RATIO), int(height / RATIO))

    def images(self) -> Tuple[Image.Image, Image.Image]:
        """Return 2 fractals docs."""
        return self.first.image(), self.second.image()

    def drop(
            self,
            data: DataExport,
            handler_progress: Optional[ProgressHandler] = None
    ) -> None:
        """Run a video from top to actual point."""
        self.first.drop(data, handler_progress)

    def swap(self) -> None:
        """Swap first with the second fractal."""
        w, h = self.first.width, self.first.height
        self.first.resize(int(w / RATIO), int(h / RATIO))
        self.second.resize(w, h)
        self.first, self.second = self.second, self.first

    def is_mandelbrot_first(self) -> bool:
        """Return if The first fractal is Mandelbrot."""
        return self.first == self.__mandelbrot

    def motion(self, x: int, y: int) -> None:
        """Set C complex of Julia."""
        if self.is_mandelbrot_first():
            r = self.__mandelbrot.real_at_x(x)
            i = self.__mandelbrot.imaginary_at_y(y)
            self.__julia.set_c_r(r)
            self.__julia.set_c_i(i)

    def save_size(self) -> int:
        """Return the size of save."""
        return 1 + self.__mandelbrot.bytes_size() + self.__julia.bytes_size()

    def save(self, path: str) -> None:
        """Save state of manager."""
        with open(path, "wb") as file:
            file.write(self.to_bytes())

    def to_bytes(self) -> bytes:
        """Convert the state of manager to bytes."""
        data = b"\x01" if self.is_mandelbrot_first() else b"\x00"
        data += self.__mandelbrot.to_bytes()
        data += self.__julia.to_bytes()
        return data

    def load(self, path: str) -> None:
        """Load fractals from path."""
        try:
            with open(path, "rb") as file:
                data = file.read(self.save_size() + 1)
        except FileNotFoundError:
            raise FileNotFoundError(
                "Le fichier {!r} n'a pas pu être trouvé".format(path)
            ) from None
        self.from_bytes(data)

    def from_bytes(self, data: bytes) -> None:
        """Load manager from bytes."""
        if len(data) != self.save_size():
            raise ValueError(
                "Mauvais type de fichier.\n"
                "Ca taille ne correspond pas au format d'un fichier .mbc")
        temp = self.to_bytes()
        try:
            is_mandelbrot = data[0]
            if is_mandelbrot != self.is_mandelbrot_first():
                self.swap()
            mandelbrot_data = data[1:36]
            julia_data = data[36:]
            self.__julia.from_bytes(julia_data)
            self.__mandelbrot.from_bytes(mandelbrot_data)
        except Exception as err:
            self.from_bytes(temp)  # need to return in previous state
            raise err

    def zoom(self, x: int, y: int, power: float) -> None:
        """Zoom in Image."""
        self.first.zoom(x, y, power)

    @property
    def pixel_size(self) -> float:
        """Get pixel size."""
        return self.first.pixel_size

    @property
    def iterations(self) -> int:
        """Get max iterations."""
        return self.first.iterations

    @iterations.setter
    def iterations(self, iterations: int) -> None:
        """Set max iterations."""
        self.first.set_iterations(iterations)
        self.second.set_iterations(iterations)

    def color(self, r: int, g: int, b: int) -> None:
        """Set RGB color."""
        color = self.__coloration
        color.r = r
        color.g = g
        color.b = b

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """Get RGB color."""
        color = self.__coloration
        return color.r, color.g, color.b

    @property
    def real(self) -> float:
        """Z Real part of Mandelbrot."""
        return self.__mandelbrot.real

    @property
    def imaginary(self) -> float:
        """Z Imaginary part of Mandelbrot."""
        return self.__mandelbrot.imaginary

    @property
    def iter_sum(self) -> int:
        """Total of all iterations."""
        return self.first.iterations_sum()

    @property
    def iter_pixel(self) -> float:
        """IterationInteraction per pixel. If not pixel return 0."""
        return self.first.iterations_per_pixel()

    @property
    def iter_second(self) -> float:
        """Iterations per seconds."""
        return 0

    def reset(self) -> None:
        """Reset fractals."""
        self.__mandelbrot.reset()
        self.__julia.reset()



# pylint: disable=unused-argument, disable=super-init-not-called, no-self-use
from typing import Optional
import numpy as np
import numpy.typing as npt
from PIL import Image
from mandelia.model.manager import ProgressHandler

from mandelia.view.export import DataExport


class ModuloColoration:
    r: int
    g: int
    b: int

    def __init__(self, r: int, g: int, b: int):
        ...

    def to_bytes(self) -> bytes:
        ...

    def from_bytes(self, data: bytes) -> None:
        ...

    def bytes_size(self) -> int:
        ...

    def colorize(
        self, np_fractale: npt.NDArray[np.uint32]
    ) -> npt.NDArray[np.uint8]:
        ...


class Fractale:
    real: float
    imaginary: float
    pixel_size: float
    width: int
    height: int
    iterations: int
    need_update: bool

    def __init__(self, color: ModuloColoration, real: float = 0,
                 imaginary: float = 0, iterations: int = 1_000,
                 width: int = 128, height: int = 128, pixel_size: float = 0.02
                 ) -> None:
        ...

    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        ...

    def __copy__(self) -> 'Fractale':
        ...

    def set_real(self, real: float) -> None:
        ...

    def set_imaginary(self, imaginary: float) -> None:
        ...

    def set_iterations(self, iterations: int) -> None:
        ...

    def resize(self, width: int, height: int) -> None:
        ...

    def iterations_sum(self) -> int:
        ...

    def iterations_per_pixel(self) -> int:
        ...

    def image(self) -> Image.Image:
        ...

    def image_at_size(self, width: int, height: int) -> Image.Image:
        ...

    def top(self) -> None:
        ...

    def middle_zoom(self, multiplier: float) -> None:
        ...

    def zoom(self, x: int, y: int, multiplier: float) -> None:
        ...

    def reset(self) -> None:
        ...

    def real_at_x(self, x: int) -> float:
        ...

    def imaginary_at_y(self, y: int) -> float:
        ...

    def to_bytes(self) -> bytes:
        ...

    def from_bytes(self, data: bytes) -> None:
        ...

    def bytes_size(self) -> int:
        ...

    def drop(
        self,
        data: DataExport,
        handler_progress: Optional[ProgressHandler] = None
    ) -> None:
        ...


class Julia(Fractale):
    c_r: float
    c_i: float

    def __init__(self, color: ModuloColoration, real: float = 0,
                 imaginary: float = 0, iterations: int = 1_000,
                 width: int = 128, height: int = 128, pixel_size: float = 0.02
                 ) -> None:
        ...

    def set_c_r(self, c_r: float) -> None:
        ...

    def set_c_i(self, c_i: float) -> None:
        ...


class Mandelbrot(Fractale):
    ...



# pylint: disable=import-outside-toplevel
# type: ignore
import os
import sys
import re
import multiprocessing

import numpy as np
from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext


class BuildExtCommand(build_ext):

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.inplace = 1


class EXECommand(BuildExtCommand):
    """A custom command to build an executable file."""

    description = "Build an executable"
    user_options = []  # type: list[str]

    def finalize_options(self):
        BuildExtCommand.finalize_options(self)

    def initialize_options(self):
        BuildExtCommand.initialize_options(self)

    def run(self):
        """Run command."""
        BuildExtCommand.run(self)

        import PyInstaller.__main__

        PyInstaller.__main__.run([
            "--noconfirm",
            "--log-level=DEBUG",
            "--clean",
            "mandelia.spec"
        ])


def read(path):
    # type: (str) -> str
    with open(path, "rt", encoding="utf8") as f:
        return f.read().strip()


def version():
    # type: () -> str
    match = re.search(r"__version__ = \"(.+)\"", read("mandelia/__init__.py"))
    if match:
        return match.group(1)
    return "0.0.1"


extra_compile_args = []  # type: list[str]
extra_link_args = []  # type: list[str]
include_dirs = [np.get_include()]  # type: list[str]
library_dirs = []  # type: list[str]
libraries = []  # type: list[str]

if sys.platform == "darwin":
    os.environ["CC"] = "/usr/local/opt/llvm/bin/clang"
    os.environ["CXX"] = "/usr/local/opt/llvm/bin/clang++"
    extra_compile_args += ["-w", "-fopenmp", "-stdlib=libc++"]
    extra_link_args += ["-fopenmp", "-stdlib=libc++"]
    include_dirs += ["/usr/local/opt/llvm/include"]
    library_dirs += ["/usr/local/opt/llvm/lib"]
elif sys.platform == "win32":
    extra_compile_args += ["/openmp"]
    extra_link_args += []
else:  # sys.platform.startswith("linux"):
    extra_compile_args += ["-fopenmp"]
    extra_link_args += ["-lgomp"]


ext = cythonize([
    Extension(
        "mandelia.model.fractale",
        ["mandelia/model/fractale.pyx"],
        libraries=libraries,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=include_dirs,
        library_dirs=library_dirs
    )],
    nthreads=multiprocessing.cpu_count(),
    annotate=True,
    language_level=3,
    build_dir="build"
)


setup(
    name='mandelia',
    version=version(),
    author="Dashstrom",
    author_email="dashstrom.pro@gmail.com",
    url='https://github.com/Dashstrom/mandelia',
    license="GPL-3.0 License",
    packages=find_packages(exclude=('tests', 'docs', '.github')),
    description="Application to visualize fractals of mandelbrot and julia.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    python_requires='>=3.6.0',
    cmdclass={
        'build_ext': BuildExtCommand,
        'exe': EXECommand
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: OS Independent",
        "Natural Language :: French"
    ],
    ext_modules=ext,
    platforms="any",
    include_package_data=True,
    test_suite="tests",
    package_data={
        "mandelia": ["view/images/*", "**/*.pyi", "**/*.pyx", "py.typed"],
    },
    keywords=["mandelbrot", "julia", "fractale", "tkinter"],
    install_requires=read("requirements.txt").split("\n"),
    entry_points={
        'console_scripts': [
            'mandelia=mandelia.__main__:main',
        ]
    },
    zip_safe=False
)



"""Controller of every view."""
import tkinter as tk
from contextlib import contextmanager
from random import randint
from tkinter.filedialog import asksaveasfilename, askopenfilename
from tkinter.messagebox import showerror, showinfo
from traceback import print_exc
from typing import Optional, Generator
from PIL import Image

from mandelia.view.export import DataExport

from ..model.manager import FractaleManager
from ..view.view import View
from ..view.wait import Wait
from ..util import logger, stat_file


class Controller:
    """General controller."""

    def __init__(self, path: Optional[str] = None) -> None:
        """Instantiate Controller."""
        with self.lock_update():
            view = self.view = View()
            view.update()
            self.manager = FractaleManager(view.width, view.height)
            self.__ignore_update = False
            self.__wait: Optional[Wait] = None

            interaction = view.interaction
            interaction.action.actualization.config(
                command=self.on_actualization)
            interaction.color.button.config(command=self.on_random_color)
            interaction.action.reset.config(command=self.on_reset)
            interaction.file.save.config(command=self.on_save)
            interaction.file.load.config(command=self.on_load)
            interaction.bind_export(self.on_export)
            interaction.iteration.max.var.trace_add(
                "write", self.on_iteration_max
            )

            view.visualization.bind("<MouseWheel>", self.on_wheel)
            view.visualization.bind("<Button-4>", self.on_right_click)
            view.visualization.bind("<Button-5>", self.on_left_click)

            view.visualization.bind("<Configure>", self.on_resize)
            view.visualization.bind('<Motion>', self.on_motion)
            view.visualization.bind('<Double-1>', self.on_swap)
            view.red.trace_add("write", lambda *_: self.on_color())
            view.green.trace_add("write", lambda *_: self.on_color())
            view.blue.trace_add("write", lambda *_: self.on_color())
        if path is not None:
            self.load_configuration(path)
        else:
            self.update()
        self.view.mainloop()

    def __repr__(self) -> str:
        """Represent a Controller."""
        name = self.__class__.__name__
        return "<{} locked_update={}>".format(name, self.locked_update)

    @logger
    def on_swap(self, event):
        # type: (tk.Event[tk.Canvas]) -> None
        """Handle swap between julia and mandelbrot."""
        self.manager.swap()
        img1, img2 = self.manager.images()
        self.view.set_image(img1)
        self.view.set_2nd_image(img2)
        self.update()

    @logger
    def on_export(self, data: DataExport) -> None:
        """Handle export."""
        if self.__wait is not None:
            showerror("Opération en cours",
                      "Un export est déjà en cours, "
                      "veuillez la fermer pour exporter de nouveau")
            return
        path = data["path"]
        ext = data["ext"]
        if ext in ("gif", "mp4"):
            wait = self.__wait = Wait(self.view)

            def handler_progress(progress: float, image: Image.Image) -> None:
                wait.progress(progress)
                wait.set_preview(image)
                if progress == 1:
                    wait.done()
                self.view.update()
        else:
            def handler_progress(progress: float, image: Image.Image) -> None:
                pass

        try:
            self.manager.drop(data, handler_progress)
            showinfo("Exporter", "Exportation réussi\n" + stat_file(path))
        except tk.TclError:
            showerror("Opération annulée", "L'opération a été annulée")
        finally:
            self.__wait = None
        self.view.update()

    def on_motion(self, event):
        # type: (tk.Event[tk.Canvas]) -> None
        """Handle mouse movement."""
        self.manager.motion(event.x, event.y)
        if self.manager.is_mandelbrot_first():
            img = self.manager.second.image()
            self.view.set_2nd_image(img)

    def load_configuration(self, path: str) -> None:
        """Load configuration with displayable error."""
        self.manager.load(path)
        self.update()
        self.update_colors()

    @contextmanager
    def lock_update(self) -> Generator[None, None, None]:
        """Context manager for lock update for a short while."""
        self.__ignore_update = True
        yield
        self.__ignore_update = False

    @property
    def locked_update(self) -> bool:
        """Return if update is locked."""
        return self.__ignore_update

    @logger
    def on_load(self) -> None:
        """Handle load of configuration."""
        path = askopenfilename(
            title="Charger la configuration",
            filetypes=[('Mandelbrot configuration Files', '.mbc')],
            initialfile="projet.mbc"
        )
        if path:
            self.load_configuration(path)

    @logger
    def on_save(self) -> None:
        """Handle save of configuration."""
        path = asksaveasfilename(
            title="Enregistrer la configuration",
            filetypes=[('Mandelbrot configuration Files', '.mbc')],
            initialfile="projet.mbc"
        )
        if path:
            try:
                self.manager.save(path)
                showinfo("Enregistrer la configuration",
                         "Enregistrement réussi\n" + stat_file(path))
            except PermissionError:
                showerror("Enregistrer la configuration",
                          "Le fichier n'a pas pu être ouvert")
            except OSError:
                showerror("Enregistrer la configuration",
                          "Une erreur est survenue")
                print_exc()

    def on_wheel(self, event):
        # type: (tk.Event[tk.Canvas]) -> None
        """Handle wheel bearing for zoom."""
        self.zoom(event.x, event.y, 2 if event.delta > 0 else 0.5)

    def on_right_click(self, event):
        # type: (tk.Event[tk.Canvas]) -> None
        """Handle right click."""
        self.zoom(event.x, event.y, 2)

    def on_left_click(self, event):
        # type: (tk.Event[tk.Canvas]) -> None
        """Handle left click."""
        self.zoom(event.x, event.y, 0.5)

    @logger
    def zoom(self, x: int, y: int, power: float) -> None:
        """Zoom in actual fractale."""
        self.manager.zoom(x, y, power)
        self.update()

    @logger
    def on_iteration_max(self, name: str, index: str, mode: str) -> None:
        """Handle change of max iterations."""
        iterations = int(self.view.interaction.iteration.max.var.get())
        self.manager.iterations = iterations

    @logger
    def on_color(self) -> None:
        """Handle color changes."""
        if self.locked_update:
            return
        self.manager.color(
            int(self.view.red.get()),
            int(self.view.green.get()),
            int(self.view.blue.get())
        )
        self.update_colors()
        self.update()

    @logger
    def update(self) -> None:
        """Update images."""
        if self.locked_update:
            return
        manager = self.manager
        view = self.view
        with self.lock_update():
            interaction = view.interaction
            interaction.positioning.real.var.set(str(manager.real))
            interaction.positioning.imaginary.var.set(str(manager.imaginary))
            interaction.positioning.zoom.var.set(str(manager.pixel_size))
            interaction.iteration.max.var.set(manager.iterations)
            interaction.iteration.sum.var.set("{} i".format(manager.iter_sum))
            interaction.iteration.per_pixel.var.set(
                "{:.2f} i/pxl".format(manager.iter_pixel))
            img = manager.first.image()
            self.view.set_image(img)

    @logger
    def on_random_color(self) -> None:
        """Handle random color button pressed."""
        rgb = [randint(0, 6) + randint(0, 6) + randint(0, 4) for _ in range(3)]
        self.manager.color(*rgb)
        self.update_colors()
        self.update()

    def update_colors(self) -> None:
        """Update colors."""
        with self.lock_update():
            r, g, b = self.manager.rgb
            self.view.red.set(r)
            self.view.green.set(g)
            self.view.blue.set(b)
            self.view.set_image(self.manager.first.image())
            self.view.set_2nd_image(self.manager.second.image())

    @logger
    def on_resize(self, event):
        # type: (tk.Event[tk.Canvas]) -> None
        """Handle window resize."""
        self.manager.resize(event.width, event.height)
        self.view.set_2nd_image(self.manager.second.image())
        self.update()

    @logger
    def on_actualization(self) -> None:
        """Handle actualization button."""
        self.manager.resize(self.view.width, self.view.height)
        self.update()

    @logger
    def on_reset(self) -> None:
        """Handle reset button."""
        self.manager.reset()
        self.view.set_2nd_image(self.manager.second.image())
        self.update()



""""Test importing specific modules."""
import sys
from unittest import TestCase


class TestImport(TestCase):

    def test_import_paramspec(self) -> None:
        # pylint: disable=import-outside-toplevel, unused-import
        if sys.version_info >= (3, 10):
            from typing import ParamSpec
        else:
            from typing_extensions import ParamSpec  # noqa: F401
        if not callable(ParamSpec):  # type: ignore
            self.fail("INvalid ParamSpec")
